;;; gnus-encrypt.el --- file encryption routines for Gnus
;; Copyright (C) 2002, 2003, 2004 Free Software Foundation, Inc.

;; Author: Teodor Zlatanov <tzz@lifelogs.com>
;; Created: 2003/01/24
;; Keywords: files

;; This file is part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;;; This module addresses data encryption under Gnus.  Page breaks are
;;; used for grouping declarations and documentation relating to each
;;; particular aspect.

;;; Code:

;; autoload password
(eval-and-compile
  (autoload 'password-read "password"))

(defgroup gnus-encrypt nil
  "Gnus encryption configuration.")

(defcustom gnus-encrypt-password-cache-expiry 20
  "Gnus encryption password timeout.
When set, directly sets password-cache-expiry"
  :type 'integer
  :group 'gnus-encrypt)

(defcustom gnus-encrypt-file-alist nil
  "List of file names or regexes matched with encryptions.
Format example:
 '((\"beta\"
    (gpg \"AES\"))
   (\"/home/tzz/alpha\"
    (gnus-encrypt-xor \"Semi-Secret\")))"

  :type '(repeat
	  (list :tag "Encryption entry"
	   (radio :tag "What to encrypt"
		  (file :tag "Filename")
		  (regexp :tag "Regular expression match"))
	   (radio :tag "How to encrypt it"
		  (list
		   :tag "GPG Encryption"
		   (const :tag "GPG Program" gpg)
		   (radio :tag "Choose a cipher"
			  (const :tag "3DES Encryption" "3DES")
			  (const :tag "CAST5 Encryption" "CAST5")
			  (const :tag "Blowfish Encryption" "BLOWFISH")
			  (const :tag "AES Encryption" "AES")
			  (const :tag "AES192 Encryption" "AES192")
			  (const :tag "AES256 Encryption" "AES256")
			  (const :tag "Twofish Encryption" "TWOFISH")
			  (string :tag "Cipher Name")))
		  (list
		   :tag "Built-in simple XOR"
		   (const :tag "XOR Encryption" gnus-encrypt-xor)
		   (string :tag "XOR Cipher Value (seed value)")))))
  :group 'gnus-encrypt)

;; TODO: now, load gencrypt.el and if successful, modify the
;; custom-type of gnus-encrypt-file-alist to add the gencrypt.el options

;; (plist-get (symbol-plist 'gnus-encrypt-file-alist) 'custom-type)
;; then use plist-put

(defcustom gnus-encrypt-gpg-path (executable-find "gpg")
  "Path to the GPG program."
  :type '(radio
	  (file :tag "Location of the GPG executable")
	  (const :tag "GPG is not installed" nil))
  :group 'gnus-encrypt)

(defvar gnus-encrypt-temp-prefix "gnus-encrypt"
  "Prefix for temporary filenames")

(defun gnus-encrypt-find-model (filename)
  "Given a filename, find a gnus-encrypt-file-alist entry"
  (dolist (entry gnus-encrypt-file-alist)
    (let ((match (nth 0 entry))
	  (model (nth 1 entry)))
      (when (or (eq match filename)
		(string-match match filename))
	(return model)))))

(defun gnus-encrypt-insert-file-contents (file)
  "Decrypt FILE into the current buffer."
  (interactive "fFile to insert: ")
  (let* ((model (gnus-encrypt-find-model file))
	 (method (nth 0 model))
	 (cipher (nth 1 model))
	 (passphrase
	  (password-read
	   (format "%s password for cipher %s? "
		   (symbol-name method) cipher)
	   (format "gnus-encrypt-password-%s-%s"
		   (symbol-name method) cipher)))
	  (buffer-file-coding-system 'binary)
	 (coding-system-for-read 'binary)
	 status-ok)
    (insert-file-contents file)

    (cond
     ((eq method 'gpg)
      (setq status-ok (gnus-encrypt-gpg-decode-buffer passphrase cipher t)))
     ((eq method 'gnus-encrypt-xor)
      (setq status-ok (gnus-encrypt-xor-decode-buffer passphrase cipher t))))

    (if status-ok
	(gnus-message 9 "%s was decrypted with %s (cipher %s)"
		      file (symbol-name method) cipher)
      (gnus-error 5 "%s was NOT decrypted with %s (cipher %s)"
		  file (symbol-name method) cipher))))

(defun gnus-encrypt-write-file-contents (file)
  "Encrypt the current buffer to FILE, then continue normally."
  (interactive "fFile to write: ")
  (let* ((model (gnus-encrypt-find-model file))
	 (method (nth 0 model))
	 (cipher (nth 1 model))
	 (contents (buffer-string))
	 (passphrase
	  (password-read
	   (format "%s password for cipher %s? "
		   (symbol-name method) cipher)
	   (format "gnus-encrypt-password-%s-%s"
		   (symbol-name method) cipher)))
	 status-ok)

    (cond
     ((eq method 'gpg)
      (setq status-ok (gnus-encrypt-gpg-encode-buffer passphrase cipher t)))
     ((eq method 'gnus-encrypt-xor)
      (setq status-ok (gnus-encrypt-xor-encode-buffer passphrase cipher t))))

    (if status-ok
	(gnus-message 9 "%s was encrypted with %s (cipher %s)"
		      file (symbol-name method) cipher)
      (gnus-error 5 "%s was NOT encrypted with %s (cipher %s)"
		  file (symbol-name method) cipher))

    ;; do not confirm overwrites
    (write-file file nil)))

;; undo-if-error is ignored
(defun gnus-encrypt-xor-encode-buffer (passphrase cipher undo-if-error)
  (gnus-encrypt-xor-process-buffer passphrase cipher t))

;; undo-if-error is ignored
(defun gnus-encrypt-xor-decode-buffer (passphrase cipher undo-if-error)
  (gnus-encrypt-xor-process-buffer passphrase cipher nil))

(defun gnus-encrypt-xor-process-buffer (passphrase
					cipher
					&optional encode)
  "Given PASSPHRASE, xor-encode or decode the contents of the current buffer."
  (let* ((bs (buffer-string))
	 ;; passphrase-sum is a simple additive checksum of the
	 ;; passphrase and the cipher
	(passphrase-sum
	 (when (stringp passphrase)
	   (apply '+ (append cipher passphrase nil))))
	new-list)
    (delete-region (point-min) (point-max))
    (goto-char (point-min))

    (if encode
	(progn
	  (dolist (x (append bs nil))
	    (setq new-list (cons (logxor x passphrase-sum) new-list)))

	  (dolist (x new-list)
	    (insert (format "%d " x))))
      (progn
	(setq new-list (reverse (split-string bs)))
	(dolist (x new-list)
	  (setq x (string-to-int x))
	  (insert (format "%c" (logxor x passphrase-sum))))))
    t))

(defun gnus-encrypt-gpg-encode-buffer (passphrase cipher undo-if-error)
  (gnus-encrypt-gpg-process-buffer passphrase cipher t undo-if-error))

(defun gnus-encrypt-gpg-decode-buffer (passphrase cipher undo-if-error)
  (gnus-encrypt-gpg-process-buffer passphrase cipher nil undo-if-error))

(defun gnus-encrypt-gpg-process-buffer (passphrase 
					cipher 
					&optional encode undo-if-error)
  "With PASSPHRASE, use GPG to encode or decode the contents of the current buffer."
  (let ((program gnus-encrypt-gpg-path)
	(old-buffer-string (buffer-string))
	(input (buffer-substring-no-properties (point-min) (point-max)))
	(temp-file (make-temp-file gnus-encrypt-temp-prefix))
	(default-enable-multibyte-characters nil)
	(args `("--cipher-algo" ,cipher
		"--status-fd" "2"
		"--logger-fd" "2"
		"--passphrase-fd" "0"
		"--no-tty"))
	exit-status)

    (when encode
      (setq args 
	    (append args
		    '("--symmetric"
		      "--armor"))))

    (if program
	(progn
	  (delete-region (point-min) (point-max))
	  (goto-char (point-min))
	  (when passphrase
	    (insert passphrase "\n"))
	  (insert input)
	  (setq exit-status
		(apply #'call-process-region (point-min) (point-max) program
		       t `(t ,temp-file) nil args))
	  (unless (equal exit-status 0)
	    (setq exit-status nil)
	    (when undo-if-error
	      (delete-region (point-min) (point-max))
	      (goto-char (point-min))
	      (insert old-buffer-string))
	    (with-temp-buffer
	      (when (file-exists-p temp-file)
		(insert-file-contents temp-file)
		(delete-file temp-file))
	      (gnus-error 5 (format "%s exited abnormally: '%s' [%s]"
				    program exit-status (buffer-string))))))
      (gnus-error 5 "GPG is not installed."))
    exit-status))

(provide 'gnus-encrypt)
;;; gnus-encrypt.el ends here
